from collections.abc import Generator
from contextlib import suppress
from typing import NamedTuple

import requirements

from labels.model.file import LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.python.package_builder import new_python_package
from labels.parsers.cataloger.utils import get_enriched_location

OPERATOR_ORDER = {"==": 1, "===": 1, "~=": 1, ">=": 2, ">": 2, "<": 3, "<=": 3}


class PythonRequirementsData(NamedTuple):
    name: str
    version: str
    is_dev: bool


def parse_requirements_txt(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = _safe_parse_requirements_txt(reader)
    if not file_content:
        return [], []

    packages = _collect_packages(file_content, reader)

    return packages, []


def _safe_parse_requirements_txt(reader: LocationReadCloser) -> str | None:
    try:
        return reader.read_closer.read()
    except UnicodeDecodeError:
        return None


def _collect_packages(file_content: str, reader: LocationReadCloser) -> list[Package]:
    packages: list[Package] = []
    deps_found = False
    for line_number, line in _split_lines_requirements(file_content):
        parsed_dependency_data = _get_parsed_dependency(line)

        if not parsed_dependency_data:
            if not deps_found and line_number > 3:
                return packages
            continue

        deps_found = True
        new_location = get_enriched_location(
            reader.location,
            line=line_number,
            is_dev=parsed_dependency_data.is_dev,
            is_transitive=False,
        )

        package = new_python_package(
            name=parsed_dependency_data.name,
            version=parsed_dependency_data.version,
            location=new_location,
        )
        if package:  # pragma: no branch
            packages.append(package)

    return packages


def _split_lines_requirements(
    content: str,
) -> Generator[tuple[int, str], None, None]:
    last_line = ""
    line_number = 1
    for index, raw_line in enumerate(content.splitlines(), 1):
        if not last_line:
            line_number = index
        line = _trim_requirements_txt_line(raw_line)
        if last_line != "":
            line = last_line + line
            last_line = ""
        if line.endswith("\\"):
            last_line += line.rstrip("\\")
            continue
        if not line:
            continue

        if any(
            (
                line.startswith("-e"),
                line.startswith("-r"),
                line.startswith("--requirements"),
            ),
        ):
            continue

        yield line_number, line


def _trim_requirements_txt_line(line: str) -> str:
    line = line.strip()

    return _remove_trailing_comment(line)


def _remove_trailing_comment(line: str) -> str:
    parts = line.split("#", 1)
    if len(parts) < 2:
        # there aren't any comments
        return line
    return parts[0]


def _get_parsed_dependency(line: str) -> PythonRequirementsData | None:
    with suppress(Exception):
        parsed_dep = next(iter(requirements.parse(line)))
        if not parsed_dep.specs or not (version := _get_dep_version_range(parsed_dep.specs)):
            return None
        is_dev: bool = False
        if parsed_dep.extras and any("dev" in extra.lower() for extra in parsed_dep.extras):
            is_dev = True
        return PythonRequirementsData(name=str(parsed_dep.name), version=version, is_dev=is_dev)
    return None


def _get_dep_version_range(dep_specs: list[tuple[str, str]]) -> str | None:
    version_range = ""
    ordered_specs = sorted(dep_specs, key=lambda x: OPERATOR_ORDER.get(x[0], 1))
    for operator, version in ordered_specs:
        if operator not in OPERATOR_ORDER:
            return None

        if operator in {"==", "~="}:
            version_range = version
            break
        version_range += f"{operator}{version} "
    return version_range.rstrip()



from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.redhat.package_builder import new_redhat_package
from labels.parsers.cataloger.redhat.rpmdb.dispatcher import RpmDB, open_db


def parse_rpm_db(
    _resolver: Resolver,
    environment: Environment,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    if not reader.location.coordinates:
        return [], []

    database = open_db(reader.location.coordinates.real_path)
    if not database:
        return [], []

    packages = _collect_packages(database, reader.location, environment)

    return packages, []


def _collect_packages(
    database: RpmDB, location: Location, environment: Environment
) -> list[Package]:
    packages: list[Package] = []

    for entry in database.list_packages():
        package = new_redhat_package(entry=entry, env=environment, location=location)
        if package is not None:
            packages.append(package)

    return packages



from more_itertools import flatten

from labels.model.file import Location, LocationReadCloser
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Package, PackageType
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.dotnet.package_builder import new_dotnet_package
from labels.parsers.cataloger.dotnet.utils import build_relationships_from_json_dependencies
from labels.parsers.cataloger.utils import get_enriched_location
from labels.parsers.collection.json import parse_json_with_tree_sitter
from labels.utils.strings import normalize_name


def parse_dotnet_deps_json(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = parse_json_with_tree_sitter(reader.read_closer.read())
    if not isinstance(file_content, IndexedDict):
        return [], []

    targets = _collect_targets(file_content)
    if not targets:
        return [], []

    packages = _collect_packages(targets, reader.location)
    relationships = _collect_relationships(targets, packages)

    return packages, relationships


def _collect_targets(file_content: IndexedDict) -> list[tuple[str, ParsedValue]] | None:
    targets: ParsedValue = file_content.get("targets")
    if not isinstance(targets, IndexedDict):
        return None

    return list(
        flatten(target.items() for target in targets.values() if isinstance(target, IndexedDict))
    )


def _collect_packages(targets: list[tuple[str, ParsedValue]], location: Location) -> list[Package]:
    packages: list[Package] = []

    for package_key, package_value in targets:
        if not isinstance(package_value, IndexedDict):
            continue

        package_info = _get_validated_package_info(package_key)
        if not package_info:
            continue

        name, version = package_info
        new_location = get_enriched_location(location, line=package_value.position.start.line)

        new_package = new_dotnet_package(name, version, new_location)
        if new_package:
            packages.append(new_package)

    return packages


def _collect_relationships(
    targets: list[tuple[str, ParsedValue]],
    packages: list[Package],
) -> list[Relationship]:
    relationships: list[Relationship] = []

    packages_by_key = {(pkg.name, pkg.version): pkg for pkg in packages}

    for package_key, package_value in targets:
        relationships.extend(
            _build_relationships_for_package(package_key, package_value, packages_by_key)
        )

    return relationships


def _build_relationships_for_package(
    package_key: str,
    package_value: ParsedValue,
    packages_by_key: dict[tuple[str, str], Package],
) -> list[Relationship]:
    relationships: list[Relationship] = []

    if not isinstance(package_value, IndexedDict):
        return relationships

    package_info = _get_validated_package_info(package_key)
    if not package_info:
        return relationships

    raw_name, version = package_info
    normalized_name = normalize_name(raw_name, PackageType.DotnetPkg)

    current_package = packages_by_key.get((normalized_name, version))
    if not isinstance(current_package, Package):
        return relationships

    dependencies = package_value.get("dependencies")
    if not isinstance(dependencies, IndexedDict):
        return relationships

    return build_relationships_from_json_dependencies(
        dependencies, current_package, packages_by_key
    )


def _get_validated_package_info(package_key: str) -> tuple[str, str] | None:
    if not _is_valid_package_key(package_key):
        return None

    return _split_package_key_into_name_and_version(package_key)


def _is_valid_package_key(package_key: str) -> bool:
    return isinstance(package_key, str) and "/" in package_key


def _split_package_key_into_name_and_version(package_key: str) -> tuple[str, str]:
    name, version = package_key.split("/", 1)
    return name, version



from labels.parsers.cataloger.redhat.rpmdb.engines.berkeley.driver import open_berkeley
from labels.parsers.cataloger.redhat.rpmdb.engines.ndb.driver import open_ndb
from labels.parsers.cataloger.redhat.rpmdb.engines.sqlite.driver import open_sqlite

__all__ = ["open_berkeley", "open_ndb", "open_sqlite"]



import re
from typing import NamedTuple

from labels.model.file import Location, LocationReadCloser
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.javascript.common import (
    DependencyMap,
    add_dependency_to_root_relationship,
    build_relationships,
    create_root_package_from_package_json,
    find_root_location,
    get_direct_dependencies_from_package_json,
    group_packages_by_name,
)
from labels.parsers.cataloger.javascript.package_builder import new_simple_npm_package
from labels.parsers.cataloger.utils import get_enriched_location
from labels.parsers.collection.yaml import parse_yaml_with_tree_sitter

VERSION_PATTERN = re.compile(r"(\d+\.\d+\.\d+(-[0-9A-Za-z\.]+)?)")


class PnpmPackageCreationDetails(NamedTuple):
    package_key: str
    package_spec: IndexedDict[str, ParsedValue]
    packages_object: IndexedDict[str, ParsedValue]
    direct_dependencies: set[str]
    base_location: Location


class PnpmPackageInfo(NamedTuple):
    name: str
    key: str
    is_dev: bool
    dependencies: set[str]


def parse_pnpm_lock(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    content = reader.read_closer.read()
    file_content = parse_yaml_with_tree_sitter(content)
    if not isinstance(file_content, IndexedDict):
        return [], []

    direct_dependencies = get_direct_dependencies_from_package_json(reader.location)
    root_package = create_root_package_from_package_json(reader.location)

    packages = _collect_packages(reader, file_content, direct_dependencies)

    if root_package:
        packages.append(root_package)

    dependency_map = _build_dependency_map_pnpm(file_content)
    key_map = _build_key_map_pnpm(file_content)

    relationships = build_relationships(packages, dependency_map, reader.location, key_map)

    if root_package:
        root_relationships = _create_root_relationships(
            root_package, packages, direct_dependencies, reader.location
        )
        relationships.extend(root_relationships)

    return packages, relationships


def _create_root_relationships(
    root_package: Package,
    packages: list[Package],
    direct_dependencies: set[str],
    location: Location,
) -> list[Relationship]:
    """Create relationships from direct dependencies to root package."""
    relationships: list[Relationship] = []

    if not (location.coordinates and location.coordinates.real_path):
        return relationships

    lockfile_path = location.coordinates.real_path
    root_location = find_root_location(root_package, lockfile_path)

    if not root_location:
        return relationships

    packages_by_name = group_packages_by_name(packages, root_package.id_)

    for dep_name in direct_dependencies:
        for dep_pkg in packages_by_name.get(dep_name, []):
            add_dependency_to_root_relationship(
                dep_pkg, root_package, root_location, lockfile_path, relationships
            )

    return relationships


def _collect_packages(
    reader: LocationReadCloser,
    file_content: IndexedDict[str, ParsedValue],
    direct_dependencies: set[str],
) -> list[Package]:
    packages: list[Package] = []

    packages_object = file_content.get("packages")
    if not isinstance(packages_object, IndexedDict):
        return packages

    for package_key, pkg_spec in packages_object.items():
        if not isinstance(pkg_spec, IndexedDict):
            continue

        package_creation_details = PnpmPackageCreationDetails(
            package_key=package_key,
            package_spec=pkg_spec,
            packages_object=packages_object,
            direct_dependencies=direct_dependencies,
            base_location=reader.location,
        )

        package = _process_package(package_creation_details)
        if package:
            packages.append(package)

    return packages


def _process_package(creation_details: PnpmPackageCreationDetails) -> Package | None:
    name_and_version = _parse_package_key(
        creation_details.package_key, creation_details.package_spec
    )
    if name_and_version is None:
        return None

    package_name, package_version = name_and_version

    is_dev = creation_details.package_spec.get("dev") is True

    package_info = PnpmPackageInfo(
        name=package_name,
        key=creation_details.package_key,
        dependencies=creation_details.direct_dependencies,
        is_dev=is_dev,
    )

    position = creation_details.packages_object.get_key_position(package_info.key)
    is_transitive = package_info.name not in package_info.dependencies

    new_location = get_enriched_location(
        creation_details.base_location,
        line=position.start.line,
        is_transitive=is_transitive,
        is_dev=package_info.is_dev,
    )

    return new_simple_npm_package(new_location, package_name, package_version)


def _parse_package_key(package: str, spec: IndexedDict[str, ParsedValue]) -> tuple[str, str] | None:
    if package.startswith("github"):
        package_name = spec.get("name")
        package_version = spec.get("version")
    else:
        package_info: list[str] = VERSION_PATTERN.split(package.strip("\"'"))
        if len(package_info) < 2:
            return None
        package_name = package_info[0].lstrip("/")[0:-1]
        package_version = package_info[1]

    if not isinstance(package_name, str) or not isinstance(package_version, str):
        return None

    return package_name, package_version


def _process_package_for_dependency_map(
    package_key: str,
    package_value: IndexedDict[str, ParsedValue],
    package_name_regex: re.Pattern[str],
    dependency_map: DependencyMap,
) -> None:
    cleaned_key = package_name_regex.sub(r"\1", package_key)
    name_and_version = _parse_package_key(cleaned_key, package_value)
    if not name_and_version:
        return

    package_name, package_version = name_and_version
    map_key = f"{package_name}@{package_version}"

    if map_key in dependency_map:
        return

    deps = package_value.get("dependencies")
    peer_deps = package_value.get("peerDependencies")
    opt_deps = package_value.get("optionalDependencies")

    dependencies_dict: dict[str, ParsedValue] = {}
    if isinstance(deps, IndexedDict):
        dependencies_dict.update(deps)
    if isinstance(peer_deps, IndexedDict):
        dependencies_dict.update(peer_deps)
    if isinstance(opt_deps, IndexedDict):
        dependencies_dict.update(opt_deps)

    if not dependencies_dict:
        return

    deps_dict = _extract_dependency_requirements(dependencies_dict)
    if deps_dict:
        dependency_map[map_key] = deps_dict


def _build_key_map_pnpm(
    package_yaml: IndexedDict[str, ParsedValue],
) -> dict[tuple[str, int], str]:
    """Build a map from (package_name, line) to package_key (name@version)."""
    key_map: dict[tuple[str, int], str] = {}

    snapshots = package_yaml.get("snapshots")
    packages_items = package_yaml.get("packages")

    sources_to_check = []
    if isinstance(snapshots, IndexedDict):
        sources_to_check.append(snapshots)
    if isinstance(packages_items, IndexedDict):
        sources_to_check.append(packages_items)

    package_name_regex = re.compile(r"^/?([^(]*)(?:\(.*\))*$")

    for source_items in sources_to_check:
        for package_key, package_value in source_items.items():
            if not isinstance(package_value, IndexedDict):
                continue

            cleaned_key = package_name_regex.sub(r"\1", package_key)
            name_and_version = _parse_package_key(cleaned_key, package_value)
            if not name_and_version:
                continue

            package_name, package_version = name_and_version
            map_key = f"{package_name}@{package_version}"

            position = source_items.get_key_position(package_key)
            key_map[(package_name, position.start.line)] = map_key

    return key_map


def _build_dependency_map_pnpm(
    package_yaml: IndexedDict[str, ParsedValue],
) -> DependencyMap:
    dependency_map: DependencyMap = {}

    snapshots = package_yaml.get("snapshots")
    packages_items = package_yaml.get("packages")

    sources_to_check = []
    if isinstance(snapshots, IndexedDict):
        sources_to_check.append(snapshots)
    if isinstance(packages_items, IndexedDict):
        sources_to_check.append(packages_items)

    package_name_regex = re.compile(r"^/?([^(]*)(?:\(.*\))*$")

    for source_items in sources_to_check:
        for package_key, package_value in source_items.items():
            if isinstance(package_value, IndexedDict):
                _process_package_for_dependency_map(
                    package_key, package_value, package_name_regex, dependency_map
                )

    return dependency_map


def _extract_dependency_requirements(
    dependencies: dict[str, ParsedValue],
) -> dict[str, str]:
    deps_dict: dict[str, str] = {}

    for raw_dep_name, raw_dep_version in dependencies.items():
        if not isinstance(raw_dep_version, str):
            continue

        # Extract package name (handles scoped packages like @babel/core)
        dep_name = _extract_package_name_from_key_dependency(raw_dep_name)
        if not dep_name:
            # If extraction fails, use the raw name (for simple packages)
            dep_name = raw_dep_name
        dep_version = _extract_version_from_value_dependency(raw_dep_version)

        # Store the version requirement
        deps_dict[dep_name] = dep_version

    return deps_dict


def _extract_package_name_from_key_dependency(item: str) -> str | None:
    # Regex pattern to extract the package name
    pattern = r"^@?[\w-]+/[\w-]+$"
    match = re.match(pattern, item)
    if match:
        return match.group(0)
    return None


def _extract_version_from_value_dependency(item: str) -> str:
    # Regex pattern to extract the version number before any parentheses
    # Includes pre-release identifiers like -beta.2, -alpha.1, -rc.3, etc.
    pattern = r"^(\d+\.\d+\.\d+(?:-[\w.]+)?)"
    match = re.match(pattern, item)
    if match:
        return match.group(1)
    return item



from typing import TypedDict

from labels.enrichers.api_interface import make_get


class OriginInfo(TypedDict):
    VCS: str
    URL: str
    Ref: str
    Hash: str


class VersionInfo(TypedDict):
    Version: str
    Time: str
    Origin: OriginInfo | None


def fetch_latest_version_info(module_path: str) -> VersionInfo | None:
    # https://go.dev/ref/mod#version-queries
    base_url = "https://proxy.golang.org"
    url = f"{base_url}/{module_path.lower()}/@latest"

    response: VersionInfo | None = make_get(url)
    return response



import re

from labels.model.ecosystem_data.debian import DpkgDBEntry
from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship, RelationshipType
from labels.model.release import Environment, Release
from labels.model.resolver import Resolver
from labels.parsers.cataloger.debian.package_builder import get_debian_package_url, new_dpkg_package

VERSION_SPECIFIERS = "[(<>="
PACKAGE_BLOCK_SEPARATOR = "\n\n"
DEPENDENCY_SEPARATOR = ","
CHOICE_SEPARATOR = "|"


def parse_dpkg_db(
    _resolver: Resolver | None,
    environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    dpkg_entries = _collect_dpkg_entries(reader.read_closer.read())

    packages = _collect_packages(dpkg_entries, reader.location, environment)
    relationships = _collect_relationships(packages)

    return packages, relationships


def _collect_dpkg_entries(file_content: str) -> list[DpkgDBEntry]:
    entries = []
    for package in file_content.strip().split(PACKAGE_BLOCK_SEPARATOR):
        lines = package.split("\n")
        data = {}
        prev_key = ""
        for line in lines:
            key, value = _handle_new_key_value(line)

            if key is not None:
                data[key] = value
                prev_key = key
            elif prev_key in data:
                data[prev_key] = f"{data[prev_key]}\n{value}"

        if all(not value for value in data.values()):
            continue

        entries.append(_parse_raw_entry(data))

    return entries


def _collect_packages(
    dpkg_entries: list[DpkgDBEntry],
    location: Location,
    environment: Environment | None,
) -> list[Package]:
    packages: list[Package] = []

    release = environment.linux_release if environment else None
    for entry in dpkg_entries:
        package = new_dpkg_package(entry, location, release)
        if package:
            packages.append(package)
            source_package = _collect_source_package(package, release)
            if source_package:
                packages.append(source_package)

    return packages


def _collect_relationships(packages: list[Package]) -> list[Relationship]:
    relationships: list[Relationship] = []
    lookup_table = _build_lookup_table(packages)

    for package in packages:
        if not isinstance(package.ecosystem_data, DpkgDBEntry):  # pragma: no cover
            continue

        ecosystem_data = package.ecosystem_data
        all_dependencies = [
            *(ecosystem_data.dependencies or []),
            *(ecosystem_data.pre_dependencies or []),
        ]

        for dep_specifier in all_dependencies:
            dependencies = _split_package_choices(dep_specifier)
            for dependency in dependencies:
                relationships.extend(
                    Relationship(
                        from_=dependency_package.id_,
                        to_=package.id_,
                        type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                    )
                    for dependency_package in lookup_table.get(dependency, [])
                )

    return relationships


def _parse_raw_entry(raw_entry: dict[str, str]) -> DpkgDBEntry:
    source_name, source_version = _extract_source_version(raw_entry.get("Source"))

    return DpkgDBEntry(
        package=raw_entry.get("Package"),
        source=source_name,
        version=raw_entry.get("Version"),
        source_version=source_version or None,
        architecture=raw_entry.get("Architecture"),
        maintainer=raw_entry.get("Maintainer"),
        provides=_split_deps(raw_entry.get("Provides")),
        dependencies=_split_deps(raw_entry.get("Depends")),
        pre_dependencies=_split_deps(raw_entry.get("Pre-Depends")),
    )


def _collect_source_package(
    package: Package,
    release: Release | None,
) -> Package | None:
    entry = package.ecosystem_data
    if not isinstance(entry, DpkgDBEntry):  # pragma: no cover
        return None

    if _should_create_source_package(entry, package):
        updated_entry: dict[str, str | None] = {
            "package": entry.source,
            "version": entry.source_version or package.version,
            "source": None,
            "source_version": None,
            "provides": None,
        }
        new_entry = entry.model_copy(update=updated_entry, deep=True)

        updated_package: dict[str, str | DpkgDBEntry | None] = {
            "name": new_entry.package,
            "version": new_entry.version,
            "p_url": get_debian_package_url(new_entry, release),
            "ecosystem_data": new_entry,
        }

        return package.model_copy(update=updated_package)

    return None


def _should_create_source_package(entry: DpkgDBEntry, package: Package) -> bool:
    has_different_source = entry.source and entry.source != package.name

    has_different_version = (
        entry.source and entry.source_version and entry.source_version != package.version
    )

    return bool(has_different_source or has_different_version)


def _build_lookup_table(packages: list[Package]) -> dict[str, list[Package]]:
    lookup: dict[str, list[Package]] = {}

    for package in packages:
        lookup.setdefault(package.name, []).append(package)

    for package in packages:
        if not isinstance(package.ecosystem_data, DpkgDBEntry):  # pragma: no cover
            continue

        ecosystem_data = package.ecosystem_data
        for provides in ecosystem_data.provides or []:
            key = _strip_version_specifier(provides)
            lookup.setdefault(key, []).append(package)

    return lookup


def _split_deps(value: str | None) -> list[str] | None:
    if not value:
        return None

    fields = value.split(DEPENDENCY_SEPARATOR)
    return [field.strip() for field in fields if field.strip()]


def _extract_source_version(source: str | None) -> tuple[str | None, str | None]:
    if not source:
        return None, None

    if match_result := re.compile(r"(?P<name>\S+)( \((?P<version>.*)\))?").match(source):
        result = match_result.groupdict()

        return result["name"], result["version"] or None

    return source, None  # pragma: no cover


def _handle_new_key_value(line: str) -> tuple[str | None, str]:
    if _is_key_value_line(line):
        key, value = line.split(":", 1)
        value = value.strip()

        return key, value

    return None, line


def _is_key_value_line(line: str) -> bool:
    return ":" in line and not line.startswith(" ")


def _strip_version_specifier(item: str) -> str:
    # Define the characters that indicate the start of a version specifier
    specifiers = VERSION_SPECIFIERS

    # Find the index of the first occurrence of any specifier character
    index = next((i for i, char in enumerate(item) if char in specifiers), None)

    # If no specifier character is found, return the original string
    if index is None:
        return item.strip()

    # Return the substring up to the first specifier character, stripped of
    # leading/trailing whitespace
    return item[:index].strip()


def _split_package_choices(value: str) -> list[str]:
    fields = value.split(CHOICE_SEPARATOR)
    return [_strip_version_specifier(field) for field in fields if field.strip()]
